#include <set>
#include <assert.h>
#include <unordered_map>

#include "circuit.h"
#include "simulator.h"
#include "paras.h"

using namespace atpg_ls;

std::string gen_name() {
    static int i = 0;
    return "E" + std::to_string(++i);
}

Gate* create_not_gate(Gate* gate) {
    Gate* not_gate = new Gate();
    not_gate->name = gen_name();
    not_gate->is_extra_gate = 1;
    not_gate->type = Gate::NOT;
    not_gate->fanins.push_back(gate);
    gate->fanouts.push_back(not_gate);
    return not_gate;
}

void Circuit::rebuild_atpg_circuit() {

    int sz = gates.size();
    assert(sz == simulator->gates.size());

    for(Gate* gate : rtopo_gates) {

        // 对于 PO 直接可以输出 SA0 = value 和 SA1 = NOT(value)
        if(gate->isPO) {
            Gate* extra_gate0 = new Gate();
            extra_gate0->type = Gate::BUF;
            extra_gate0->fanins.push_back(gate);
            gate->fanouts.push_back(extra_gate0);
            extra_gate0->is_extra_gate = 1;
            gates.push_back(extra_gate0);
            extra_gate0->name = "SA0_" + gate->name;

            Gate* extra_gate1 = new Gate();
            extra_gate1->type = Gate::NOT;
            extra_gate1->fanins.push_back(gate);
            gate->fanouts.push_back(extra_gate1);
            extra_gate1->is_extra_gate = 1;
            gates.push_back(extra_gate1);
            extra_gate1->name = "SA1_" + gate->name;

            // gate->isPO = false;

            gate->stuck_at_gate[0] = extra_gate0;
            gate->stuck_at_gate[1] = extra_gate1;
            continue;
        }

        // create a gate for SA_i

        std::vector<Gate*> or_inputs;

        for(Gate *out : gate->fanouts) {

            if(out->is_extra_gate) continue;

            // if the fault SA(!out->value) can be detected
            Gate* out_detected = new Gate();
            out_detected->name = gen_name();
            out_detected->is_extra_gate = 1;
            out_detected->type = Gate::OR;
            out_detected->fanins.push_back(out->stuck_at_gate[0]);
            out_detected->fanins.push_back(out->stuck_at_gate[1]);

            assert(out->stuck_at_gate[0]);
            assert(out->stuck_at_gate[1]);

            out->stuck_at_gate[0]->fanouts.push_back(out_detected);
            out->stuck_at_gate[1]->fanouts.push_back(out_detected);

            gates.push_back(out_detected);

            if(out->type == Gate::XOR || out->type == Gate::XNOR || out->type == Gate::NOT || out->type == Gate::BUF || out->type == Gate::LINE) {
                or_inputs.push_back(out_detected);
                continue;
            }

            Gate *extra_gate = new Gate();
            extra_gate->is_extra_gate = 1;
            extra_gate->type = Gate::AND;
            extra_gate->name = gen_name();
            gates.push_back(extra_gate);
            

            for(Gate *in_of_out : out->fanins) {
                if(in_of_out == gate) continue;

                if(out->type == Gate::AND || out->type == Gate::NAND) {

                    in_of_out->fanouts.push_back(extra_gate);
                    extra_gate->fanins.push_back(in_of_out);

                } else if(out->type == Gate::OR || out->type == Gate::NOR) {

                    Gate* not_gate = create_not_gate(in_of_out);
                    not_gate->name = gen_name();
                    gates.push_back(not_gate);
                    not_gate->fanouts.push_back(extra_gate);
                    extra_gate->fanins.push_back(not_gate);

                } else {
                    assert(0);
                }
            }

            extra_gate->fanins.push_back(out_detected);
            out_detected->fanouts.push_back(extra_gate);

            or_inputs.push_back(extra_gate);
        }

        Gate* or_gate;
        if(or_inputs.size() == 1) {
            or_gate = or_inputs[0];
        } else {
            or_gate = new Gate();
            gates.push_back(or_gate);
            or_gate->name = gen_name();
            or_gate->is_extra_gate = 1;
            or_gate->type = Gate::OR;
            for(Gate* in : or_inputs) {
                or_gate->fanins.push_back(in);
                in->fanouts.push_back(or_gate);
            }
        }

        Gate* p_sa0 = new Gate();
        gates.push_back(p_sa0);
        p_sa0->is_extra_gate = 1;
        p_sa0->type = Gate::AND;
        p_sa0->fanins.push_back(or_gate);
        p_sa0->fanins.push_back(gate);
        or_gate->fanouts.push_back(p_sa0);
        gate->fanouts.push_back(p_sa0);

        Gate* p_sa1 = new Gate();
        gates.push_back(p_sa1);
        p_sa1->is_extra_gate = 1;
        p_sa1->type = Gate::AND;
        p_sa1->fanins.push_back(or_gate);
        or_gate->fanouts.push_back(p_sa1);

        Gate* not_gate = create_not_gate(gate);
        not_gate->name = gen_name();
        gates.push_back(not_gate);

        p_sa1->fanins.push_back(not_gate);
        not_gate->fanouts.push_back(p_sa1);

        gate->stuck_at_gate[0] = p_sa0;
        gate->stuck_at_gate[1] = p_sa1;

        p_sa0->name = "SA0_" + gate->name;
        p_sa1->name = "SA1_" + gate->name;
    }

    // for(Gate* po : POs) {
    //     po->isPO = false;
    // }
    // POs.clear();

    for(Gate* g : gates) {
        if(g->is_extra_gate) continue;

        assert(g->stuck_at_gate[0]);
        assert(g->stuck_at_gate[1]);

        g->stuck_at_gate[0]->isPO = 1;
        g->stuck_at_gate[1]->isPO = 1;

        POs.push_back(g->stuck_at_gate[0]);
        POs.push_back(g->stuck_at_gate[1]);

    }

    init_topo_index();

    for(Gate* pi : PIs) {
        pi->value = rand() % 2;
    }

    for(Gate* g : gates) {
        g->value = g->cal_value();
    }

    for(Gate* g : rtopo_gates) {
        if(g->is_extra_gate) continue;

        g->fault_detected[0] = g->cal_fault_detected(0);
        g->fault_detected[1] = g->cal_fault_detected(1);

        assert(g->fault_detected[0] == g->stuck_at_gate[0]->value);
        assert(g->fault_detected[1] == g->stuck_at_gate[1]->value);
    }

    for(Gate* g : gates) {
        if(g->is_extra_gate) continue;
        original_gates.push_back(g);
    }

    for(int i=0; i<original_gates.size(); i++) {
        original_gates[i]->index = i;
        simulator->name2gate[original_gates[i]->name]->index = i;
    }

    std::vector<Gate*> t_POs;

    for(Gate* po : POs) {
        if(po->is_extra_gate) {
            t_POs.push_back(po);
        }
    }

    POs = t_POs;
    assert(original_gates.size() * 2 == POs.size());
    assert(original_gates.size() == sz);
}

void Circuit::insert_lines_for_stem() {

    for(int j=0; j<gates.size(); j++) {

        Gate* g = gates[j];

        if(g->type == Gate::LINE) break;

        g->is_stem = g->isPO ? g->fanouts.size() >= 1 : g->fanouts.size() >= 2;
        if(!g->is_stem) continue;

        if(g->isPO) {
            Gate* line = new Gate();
            line->name = g->name + "_line_PO";
            line->type = Gate::LINE;
            line->isPI = false;
            line->isPO = true;
            line->is_stem = false;

            line->fanins.push_back(g);
            g->fanouts.push_back(line);

            g->isPO = false;
            POs.erase(std::find(POs.begin(), POs.end(), g));
            POs.push_back(line);
            
            gates.push_back(line);
            name2gate.insert(std::make_pair(line->name, line));
        }

        // printf(">>>> now: %s\n", g->name.c_str());

        // printf("outs: [ ");
        // for(Gate* out :g->fanouts){
        //     printf("%s ", out->name.c_str());
        // }
        // printf("]\n");

        for(int i=0; i<g->fanouts.size(); i++) {
            Gate* out = g->fanouts[i];

            if(out->type == Gate::LINE) break;

            // printf("        g_name: %s  outname: %s\n", g->name.c_str(), out->name.c_str());

            Gate* line = new Gate();
            line->name = g->name + "_line_" + out->name;
            line->type = Gate::LINE;
            line->isPI = false;
            line->isPO = false;
            line->is_stem = false;

            line->fanins.push_back(g);
            line->fanouts.push_back(out);

            out->fanins.erase(std::find(out->fanins.begin(), out->fanins.end(), g));
            out->fanins.push_back(line);

            g->fanouts.erase(std::find(g->fanouts.begin(), g->fanouts.end(), out));
            g->fanouts.push_back(line);

            gates.push_back(line);
            name2gate.insert(std::make_pair(line->name, line));

            i--;
        }

    }

}

void Circuit::init_topo_index() {
    int topo = 1;
    std::queue<Gate*> q;

    std::unordered_map<Gate*, int> ins;
    for(Gate* gate : gates) {
        ins[gate] = gate->fanins.size();
    }

    for(auto in : PIs) {
        in->id = topo++;
        q.push(in);
    }

    while(!q.empty()) {
        Gate* g = q.front(); q.pop();
        for(Gate* out : g->fanouts) {
            ins[out]--;
            if(ins[out] == 0) {
                out->id = topo++;
                q.push(out);
            }
        }
    }

    rtopo_gates.clear();

    // 计算反向拓扑序
    topo = 1;
    std::unordered_map<Gate*, int> outs;

    for(Gate* g : gates) {
        if(g->fanouts.size() == 0) {
            g->rtopo = topo++;
            q.push(g);
        }
    }

    for(Gate* gate : gates) {
        outs[gate] = gate->fanouts.size();
    }

    std::unordered_set<Gate*> ok;

    while(!q.empty()) {
        Gate* g = q.front(); q.pop();
        rtopo_gates.push_back(g);
        ok.insert(g);
        for(Gate* in : g->fanins) {
            outs[in]--;

            if(outs[in] == 0) {
                in->rtopo = topo++;
                q.push(in);
            }
        }
    }

    assert(gates.size() == rtopo_gates.size());

    std::sort(gates.begin(), gates.end(), [](Gate* a, Gate* b) {
        return a->id < b->id;
    });
}


void Circuit::print() {

    printf("PIs: ( ");
    for(Gate* gate : PIs) {
        printf("%s ", gate->name.c_str());
    }
    printf(")\n");

    const char *name[10] = {"AND", "NAND", "OR", "NOR", "XOR", "XNOR", "NOT", "BUF", "INPUT", "LINE"};

    

    for(Gate* g : gates) {
        std::string out_name = "unknown";
        if(g->name != "") {
            out_name = g->name;
        }
        printf("[sa0: %d sa1: %d v: %d vsat: %d] %s = %s (",g->fault_detected[0], g->fault_detected[1], g->value, g->cal_value() == g->value,  out_name.c_str(), name[g->type]);
        for(Gate* in : g->fanins) {
            printf("%s ", in->name.c_str());
        }
        printf(")\n");
    }

    printf("POs: ( ");
    for(Gate* gate : POs) {
        printf("%s ", gate->name.c_str());
    }
    printf(")\n");
}



void Circuit::init_faults() {

    int sz = gates.size();
    assert(sz == simulator->gates.size());
    for(int i=0; i<sz; i++) {
        assert(gates[i]->name == simulator->gates[i]->name);
        simulator->gates[i]->ori_gate = gates[i];
        gates[i]->sim_gate = simulator->gates[i];
    }

    std::vector<TMP_FAULT> faults;

    for(Gate* g : gates) {

        std::string name = g->name;

        if(g->is_stem) {
            faults.push_back(TMP_FAULT{g, name, 0, 1, 0});
            faults.push_back(TMP_FAULT{g, name, 1, 1, 0});
        } else {

            if(g->isPO) {
                faults.push_back(TMP_FAULT{g, name, 0, g->type != Gate::LINE, g->type == Gate::LINE});
                faults.push_back(TMP_FAULT{g, name, 1, g->type != Gate::LINE, g->type == Gate::LINE});
                continue;
            }

            assert(g->fanouts.size() == 1);
            
            Gate *fanout = g->fanouts[0];

            int stem = (!g->isPI) && (g->type != Gate::LINE);

            if(fanout->type == Gate::Type::BUF || fanout->type == Gate::Type::NOT) {
                continue;
            } else if(fanout->type == Gate::Type::XOR || fanout->type == Gate::Type::XNOR) {
                faults.push_back(TMP_FAULT{g, name, 0, stem, 0});
                faults.push_back(TMP_FAULT{g, name, 1, stem, 0});
            } else if(fanout->type == Gate::Type::NAND || fanout->type == Gate::Type::AND) {
                faults.push_back(TMP_FAULT{g, name, 1, stem, 0});
            } else if(fanout->type == Gate::Type::NOR || fanout->type == Gate::Type::OR) {
                faults.push_back(TMP_FAULT{g, name, 0, stem, 0});
            } else {
                assert(false);
            }
        }
    }

    printf("fault-size: %d\n", faults.size());

    for(TMP_FAULT &f : faults) {
        mp[Fault{f.g, f.stuck_at}] = f;
        this->faults.insert(Fault{f.g, f.stuck_at});
    }

    this->unknown_faults = this->faults;

    assert(faults.size() == mp.size());
}