#include <set>
#include <chrono>
#include <iostream>
#include <map>

#include "circuit.h"
#include "paras.h"
#include "simulator.h"
#include "sat_atpg.h"
#include "pattern.h"
#include "checker.h"

using namespace atpg_ls;

std::set<Gate*> unjust_gates;

// score > 0 && CC = 1;
std::vector<Gate*> good_gates;

int Circuit::cal_circuit_score() {
    int score = 0;

    for(Gate* g : gates) {
        if(g->value == 1) {
            score += g->weight;
        }

        assert(g->v_sat == (g->value == g->cal_value()));

        if(!g->v_sat) {
            score -= g->unsat_cost;
        }

        if(g->v_sat) {
            assert(unjust_gates.count(g) == 0);
        } else {
            assert(unjust_gates.count(g));
        }
    }

    return score;
}

void Circuit::ls_init() {
    unjust_gates.clear();
    good_gates.clear();
}

void Circuit::ls_random_sol() {
    for(Gate* gate : gates) {
        gate->value = rand() % 2;
    }
    for(Gate* gate : gates) {
        gate->v_sat = (gate->value == gate->cal_value());
        gate->weight = 0;
        gate->CC = 1;
        gate->unsat_cost = 1;
        gate->is_good_var = 0;
    }

    for(Fault f : unknown_faults) {
        Gate* g_sai = f.gate->stuck_at_gate[f.stuck_at];
        assert(g_sai != nullptr);
        g_sai->weight = 1;
    }
}

Gate* Circuit::ls_pick() {
    Gate* pick = nullptr;

    if(!good_gates.empty()) {
        pick = good_gates[rand() % good_gates.size()];
        if(pick->score > 0 && pick->CC) {
            return pick;
        } else {
            std::vector<Gate*> tmp = good_gates;
            good_gates.clear();
            for(Gate* gate : tmp) {
                if(gate->score > 0 && gate->CC == 1) {
                    good_gates.push_back(gate);
                } else {
                    gate->is_good_var = 0;
                }
            }

            if(!good_gates.empty()) {
                pick = good_gates[rand() % good_gates.size()];
                return pick;
            }
        }
    }

    ls_update();

    std::vector<Gate*> uj;
    uj.insert(uj.end(), unjust_gates.begin(), unjust_gates.end());
     
    if(!unjust_gates.empty()) {
        pick = uj[rand() % uj.size()];

        if(rand() % 10000 <= 0.7 * 10000) {
            return pick->fanins[rand() % pick->fanins.size()];
        } else {
            return pick;
        }
    }

    pick = gates[rand() % gates.size()];

    return pick;
}

void Circuit::ls_flip(Gate *gate) {
    // for(Gate* g : gates) {
    //     int old = g->score;
    //     g->cal_score();
    //     assert(old == g->score);
    // }

    gate->value ^= 1;
    gate->CC = 0;
    gate->is_good_var = 0;

    if(!gate->isPI) {

        gate->v_sat ^= 1;
        assert(gate->v_sat == (gate->value == gate->cal_value()));
        // printf("test: %d %d\n", gate->v_sat, (gate->value == gate->cal_value()));

        if(gate->v_sat) {
            unjust_gates.erase(gate);
        } else {
            unjust_gates.insert(gate);
        }
    }

    for(Gate* in : gate->fanins) {
        in->cal_score();
        in->CC = 1;
        if(!in->is_good_var && in->score > 0) {
            in->is_good_var = 1;
            good_gates.push_back(in);
        }
    }

    for(Gate* out : gate->fanouts) {

        if(out->v_sat && (out->value != out->cal_value())) {
            out->v_sat = 0;
            unjust_gates.insert(out);
        }
        if(!out->v_sat && (out->value == out->cal_value())) {
            out->v_sat = 1;
            unjust_gates.erase(out);
        }

        out->cal_score();
        out->CC = 1;
        if(!out->is_good_var && out->score > 0) {
            out->is_good_var = 1;
            good_gates.push_back(out);
        }

        for(Gate* in : out->fanins) {
            if(in == gate) continue;
            in->CC = 1;
            in->cal_score();
            if(!in->is_good_var && in->score > 0) {
                in->is_good_var = 1;
                good_gates.push_back(in);
            }
        }
    }

    gate->cal_score();
}

void Circuit::ls_update() {

    std::set<Gate*> reigon;

    for(Gate* g : unjust_gates) {
        g->unsat_cost += OPT(vsat_inc);

        for(Gate* in : g->fanins) {
            reigon.insert(in);
        }
        for(Gate* out : g->fanouts) {
            reigon.insert(out);
            for(Gate* in : out->fanins) {
                reigon.insert(in);
            }
        }
    }

    for(Fault f : unknown_faults) {
        Gate* g = f.gate->stuck_at_gate[f.stuck_at];
        assert(g != nullptr);
        g->weight += OPT(fw_inc) * 100;
        for(Gate* in : g->fanins) {
            reigon.insert(in);
        }
        for(Gate* out : g->fanouts) {
            reigon.insert(out);
            for(Gate* in : out->fanins) {
                reigon.insert(in);
            }
        }
    }

    for(Gate* g : reigon) {
        g->cal_score();
        if(!g->is_good_var && g->score > 0 && g->CC) {
            g->is_good_var = 1;
            good_gates.push_back(g);
        }
    }
}

void Circuit::ls_gen_sol(const Fault &f) {

    std::vector<int> best_sol;
    int best_score = -1;
    simulator->simulate(PIs, best_score, unknown_faults);
    for(Gate* gate : PIs) {
        best_sol.push_back(gate->value);
    }

    for(Gate* gate : gates) {
        gate->value = gate->cal_value();
        gate->v_sat = 1;
        gate->CC = 1;
        gate->is_good_var = 0;
    }

    for(Gate* gate : gates) {
        gate->cal_score();
        if(gate->score > 0) {
            gate->is_good_var = 1;
            good_gates.push_back(gate);
        }
    }

    
    for(int step=0; step<OPT(max_step_coeff)*original_gates.size(); step++) {

        // assert(unjust_gates.size() == 0);
        // printf("step: %d\n", step);

        Gate* pick = ls_pick();

        // int score1 = cal_circuit_score();
        
        pick->cal_score();
        int delta = pick->score;
        // printf("flip: %s score: %d\n", pick->name.c_str(), delta);

        ls_flip(pick);

        // int score2 = cal_circuit_score();

        // printf("score: %d %d %d\n", score1, score2, delta);
        // assert((score2 - score1) == delta);
    
        if(pick->isPI) {
        
            Pattern p;
            for(Gate* gate : PIs) {
                p.input_vector.push_back(gate->value);
            }

            int res = check_fault(p, f);

            if(!res) {
                ls_flip(pick);
                pick->CC = 0;
            } else {

                printf("succsss-flip-pi: %s\n", pick->name.c_str());
                int score;
                simulator->simulate(PIs, score, unknown_faults);
                if(score > best_score) {
                    printf("update_best_score: %d\n", best_score);
                    best_score = score;
                    best_sol.clear();
                    for(Gate* gate : PIs) {
                        p.input_vector.push_back(gate->value);
                    }
                }
            }
        }
    }

    // printf("====== test end\n");
    // exit(0);

    assert(best_score != -1);

    for(int i=0; i<PIs.size(); i++) {
        PIs[i]->value = best_sol[i];
    }
}

void Circuit::ls_main() {

    auto start = std::chrono::high_resolution_clock::now();

    printf("====== local search start ======\n");

    std::vector<Pattern> patterns;

    while(!unknown_faults.empty()) {

        Fault f = *unknown_faults.begin(); unknown_faults.erase(unknown_faults.begin());

        ls_init();

        ls_random_sol();

        Gate* g = f.gate;
        int stuck_at = f.stuck_at;

        printf("start with fault: %s SA%d\t", g->name.c_str(), stuck_at);

        printf("verify ...");

        std::vector<int> inputs;

        int res1 = sat_atpg(mp[f], inputs);
        
        if(res1 == 0) {
            printf(" unsat!\n");
            undetected_faults.insert(f);
            continue;
        }

        printf("sat !\n");

        assert(inputs.size() == PIs.size());

        for(int i=0; i<inputs.size(); i++) {
            PIs[i]->value = inputs[i];
        }

        ls_gen_sol(f);
        
        int score;
        simulator->simulate(PIs, score, unknown_faults);

        Pattern pattern;
        for(Gate* g : PIs) {
            pattern.input_vector.push_back(g->value);
        }

        assert(check_fault(pattern, f));

        detected_faults.insert(f);
        pattern.detected_faults.push_back(f);

        int num_fault = 1;

        std::vector<Fault> remove_faults;

        for(Fault f : unknown_faults) {

            Gate* gate = f.gate;
            int stuck_at = f.stuck_at;

            assert(gate->sim_gate != nullptr);

            if(gate->sim_gate->fault_detected[stuck_at]) {

                int res = check_fault(pattern, f);

                if(!res) {
                    printf("fault: %s SA%d is not real detected!\n", gate->name.c_str(), stuck_at);
                } else {
                    pattern.detected_faults.push_back(f);
                    num_fault++;
                    detected_faults.insert(f);
                    remove_faults.push_back(f);
                }
            }
        }

        for(Fault f : remove_faults) {
            unknown_faults.erase(f);
        }

        patterns.push_back(pattern);

        auto end = std::chrono::high_resolution_clock::now();
        auto duration = std::chrono::duration_cast<std::chrono::milliseconds>(end - start).count();

        printf("Cover: %.2f%% pattern: %d new_detected: %d undected: %d time: %.2fs\n", (double)detected_faults.size() / faults.size() * 100, patterns.size(), num_fault, unknown_faults.size(), (double)duration/1000);

    }

    printf("====== local search end ======\n");

    final_check(faults, patterns);

    auto end = std::chrono::high_resolution_clock::now();
    auto duration = std::chrono::duration_cast<std::chrono::milliseconds>(end - start).count();
    std::cout << "Execution time: " << duration << " milliseconds" << std::endl;
}